# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Gridded PSF models.
"""

import copy
import itertools

import numpy as np
from astropy.io import registry
from astropy.modeling import Fittable2DModel, Parameter
from astropy.nddata import NDData
from astropy.utils.decorators import lazyproperty
from scipy.interpolate import RectBivariateSpline

from photutils.psf.model_io import (GriddedPSFModelRead, _get_metadata,
                                    _read_stdpsf, is_stdpsf, is_webbpsf,
                                    stdpsf_reader, webbpsf_reader)
from photutils.psf.model_plotting import (_ModelGridPlotter,
                                          _plot_grid_docstring)
from photutils.utils._parameters import as_pair

__all__ = ['GriddedPSFModel', 'STDPSFGrid']
__doctest_skip__ = ['STDPSFGrid']


class GriddedPSFModel(Fittable2DModel):
    """
    A model for a grid of 2D ePSF models.

    The ePSF models are defined at fiducial detector locations and are
    (x, y) detector position. The fiducial detector locations are must
    form a rectangular grid.

    The model has three model parameters: an image intensity scaling
    factor (``flux``) which is applied to the input image, and two
    positional parameters (``x_0`` and ``y_0``) indicating the location
    of a feature in the coordinate grid on which the model is evaluated.

    When evaluating this model, it cannot be called with x and y arrays
    that have greater than 2 dimensions.

    Parameters
    ----------
    nddata : `~astropy.nddata.NDData`
        A `~astropy.nddata.NDData` object containing the grid of
        reference ePSF arrays. The data attribute must contain a 3D
        `~numpy.ndarray` containing a stack of the 2D ePSFs with a shape
        of ``(N_psf, ePSF_ny, ePSF_nx)``. The length of the x and y
        axes must both be at least 4. ``N_psf`` must not be 2 or 3. All
        elements of the input image data must be finite. The PSF peak is
        assumed to be located at the center of the input image. Please
        see the Notes section below for details on the normalization of
        the input image data.

        If ``N_psf`` is 1, the model will be evaluated using the single
        ePSF image at every (x, y) position. This is equivalent to using
        the `~photutils.psf.ImagePSF` model with the single ePSF.

        The meta attribute must be dictionary containing the following:

        * ``'grid_xypos'``: A list of the (x, y) grid positions of
          each reference ePSF. The order of positions should match the
          first axis of the 3D `~numpy.ndarray` of ePSFs. In other words,
          ``grid_xypos[i]`` should be the (x, y) position of the reference
          ePSF defined in ``nddata.data[i]``. The grid positions must form
          a rectangular grid.

        * ``'oversampling'``: The integer oversampling factor(s) of the
          input ePSF images. If ``oversampling`` is a scalar then it will
          be used for both axes. If ``oversampling`` has two elements,
          they must be in ``(y, x)`` order.

        The meta attribute may contain other properties such as the
        telescope, instrument, detector, and filter of the ePSF.

    flux : float, optional
        The flux scaling factor for the model. This is the total flux
        of the source, assuming the input ePSF images are properly
        normalized.

    x_0, y_0 : float, optional
        The (x, y) position of the PSF peak in the image in the output
        coordinate grid on which the model is evaluated.

    fill_value : float, optional
        The value to use for points outside the input pixel grid. The
        default is 0.0.

    Methods
    -------
    read(*args, **kwargs)
        Class method to create a `GriddedPSFModel`
        instance from a STDPSF FITS file. This method uses
        :func:`~photutils.psf.stdpsf_reader` with the provided
        parameters.

    Notes
    -----
    The fitted PSF model flux represents the total flux of the source,
    assuming the input image was properly normalized. This flux is
    determined as a multiplicative scale factor applied to the input
    image PSF, after accounting for any oversampling. Theoretically,
    the sum of all values in the PSF image over an infinite grid should
    equal 1.0 (assuming no oversampling). However, when the PSF is
    represented over a finite region, the sum of the values may be less
    than 1.0. For oversampled PSF images, the normalization should be
    adjusted so that the sum of the array values equals the product
    of the oversampling factors (e.g., oversampling squared if the
    oversampling is the same along both axes). If the input image only
    covers a finite region of the PSF, the sum may again be less than
    the product of the oversampling factors. Correction factors based on
    the encircled or ensquared energy of the PSF can be used to estimate
    the proper scaling for the finite region of the input PSF image and
    ensure correct flux normalization.

    Internally, the grid of ePSFs will be arranged and stored such that
    it is sorted first by the y reference pixel coordinate and then by
    the x reference pixel coordinate.
    """

    flux = Parameter(description='Intensity scaling factor for the ePSF '
                     'model.', default=1.0)
    x_0 = Parameter(description='x position in the output coordinate grid '
                    'where the model is evaluated.', default=0.0)
    y_0 = Parameter(description='y position in the output coordinate grid '
                    'where the model is evaluated.', default=0.0)

    read = registry.UnifiedReadWriteMethod(GriddedPSFModelRead)

    def __init__(self, nddata, *, flux=flux.default, x_0=x_0.default,
                 y_0=y_0.default, fill_value=0.0):

        self._data, self._grid_xypos = self._define_grid(nddata)
        self._meta = nddata.meta.copy()  # _meta to avoid the meta descriptor
        self._oversampling = as_pair('oversampling',
                                     nddata.meta['oversampling'],
                                     lower_bound=(0, 0))
        self.fill_value = fill_value

        self._xgrid = np.unique(self.grid_xypos[:, 0])  # sorted
        self._ygrid = np.unique(self.grid_xypos[:, 1])  # sorted
        self.meta['grid_shape'] = (len(self._ygrid), len(self._xgrid))

        self._interpolator = {}

        super().__init__(flux, x_0, y_0)

    @staticmethod
    def _validate_data(data):
        """
        Validate the input ePSF data.

        Parameters
        ----------
        data : `~astropy.nddata.NDData`
            The input NDData object containing the ePSF data.

        Raises
        ------
        TypeError
            If the input data is not an NDData instance.
        ValueError
            If the input data is not a 3D numpy ndarray or if the input
            data contains NaNs or infs.
        """
        if not isinstance(data, NDData):
            msg = 'data must be an NDData instance'
            raise TypeError(msg)

        if data.data.ndim != 3:
            msg = 'The NDData data attribute must be a 3D numpy ndarray'
            raise ValueError(msg)

        if not np.all(np.isfinite(data.data)):
            msg = 'All elements of input data must be finite'
            raise ValueError(msg)

        if data.data.shape[0] in (2, 3):
            msg = 'The number of ePSFs must not be 2 or 3'
            raise ValueError(msg)

        # this is required by RectBivariateSpline for kx=3, ky=3
        if np.any(np.array(data.data.shape[1:]) < 4):
            msg = ('The length of the PSF x and y axes must both be at '
                   'least 4')
            raise ValueError(msg)

        if 'oversampling' not in data.meta:
            msg = "'oversampling' must be in the nddata meta dictionary"
            raise ValueError(msg)

    @staticmethod
    def _is_rectangular_grid(grid_xypos):
        """
        Determine if the given (x, y) pixel positions form an axis-
        aligned rectangular grid.

        Spacing does not need to be uniform along x or y, but there must
        be at least two unique x and y values, and all combinations of x
        and y must be present.

        Parameters
        ----------
        grid_xypos : (N, 2) array of (x, y) pairs
            The fiducial (x, y) positions of the ePSFs.

        Returns
        -------
        result : bool
            Returns `True` if the input ``grid_xypos`` forms a
            rectangular grid.
        """
        if len(grid_xypos) < 4:  # pragma: no cover
            return False

        x_vals = np.unique(grid_xypos[:, 0])  # sorted
        y_vals = np.unique(grid_xypos[:, 1])  # sorted
        # Must have at least 2 unique x and y values to form a 2D grid
        if len(x_vals) < 2 or len(y_vals) < 2:
            return False

        expected_points = {(x, y) for x in x_vals for y in y_vals}
        return set(map(tuple, grid_xypos)) == expected_points

    def _validate_grid(self, data):
        """
        Validate the input ePSF grid.

        Parameters
        ----------
        data : `~astropy.nddata.NDData`
            The input NDData object containing the ePSF data.

        Raises
        ------
        ValueError
            If the input grid_xypos does not form a rectangular grid.
        """
        try:
            grid_xypos = np.array(data.meta['grid_xypos'])
        except KeyError as exc:
            msg = "'grid_xypos' must be in the nddata meta dictionary"
            raise ValueError(msg) from exc

        if len(grid_xypos) != data.data.shape[0]:
            msg = ('The length of grid_xypos must match the number of '
                   'input ePSFs')
            raise ValueError(msg)

        if len(grid_xypos) != 1 and not self._is_rectangular_grid(grid_xypos):
            msg = 'grid_xypos must form a rectangular grid'
            raise ValueError(msg)

    def _define_grid(self, nddata):
        """
        Sort the input ePSF data into a rectangular grid where the ePSFs
        are sorted first by y and then by x.

        Parameters
        ----------
        nddata : `~astropy.nddata.NDData`
            The input NDData object containing the ePSF data.

        Returns
        -------
        data : 3D `~numpy.ndarray`
            The 3D array of ePSFs.
        grid_xypos : array of (x, y) pairs
            The (x, y) positions of the ePSFs, sorted first by y and
            then by x.
        """
        self._validate_data(nddata)
        self._validate_grid(nddata)

        grid_xypos = np.array(nddata.meta['grid_xypos'])
        # sort by y and then by x (last key is primary)
        idx = np.lexsort((grid_xypos[:, 0], grid_xypos[:, 1]))
        return nddata.data[idx], grid_xypos[idx]

    def __str__(self):
        keywords = []

        keys = ('STDPSF', 'instrument', 'detector', 'filter')
        for key in keys:
            if key in self.meta:
                name = key.capitalize() if key != 'STDPSF' else key
                keywords.append((name, self.meta[key]))

        keywords.extend([('Number of PSFs', len(self.grid_xypos)),
                         ('Grid shape', self.meta['grid_shape']),
                         ('Grid positions', self.grid_xypos.tolist()),
                         ('PSF shape (oversampled pixels)',
                          self.data.shape[1:]),
                         ('Oversampling', self.oversampling.tolist()),
                         ('Fill Value', self.fill_value)])

        return self._format_str(keywords=keywords)

    def __repr__(self):
        kwargs = {'oversampling': self.oversampling.tolist(),
                  'fill_value': self.fill_value}
        return self._format_repr(args=[], kwargs=kwargs)

    @property
    def data(self):
        """
        The 3D array of ePSFs.

        The shape is ``(N_psf, ePSF_ny, ePSF_nx)``.
        """
        return self._data

    @property
    def grid_xypos(self):
        """
        The (x, y) positions of the ePSFs.

        The order of positions should match the first axis of the 3D
        `~numpy.ndarray` of ePSFs. In other words, ``grid_xypos[i]``
        should be the (x, y) position of the reference ePSF defined in
        ``nddata.data[i]``. The grid positions must form a rectangular
        grid.
        """
        return self._grid_xypos

    def copy(self):
        """
        Return a copy of this model where only the model parameters are
        copied.

        All other copied model attributes are references to the
        original model. This prevents copying the ePSF grid data, which
        may contain a large array.

        This method is useful if one is interested in only changing
        the model parameters in a model copy. It is used in the PSF
        photometry classes during model fitting.

        Use the `deepcopy` method if you want to copy all the model
        attributes, including the ePSF grid data.

        Returns
        -------
        result : `GriddedPSFModel`
            A copy of this model with only the model parameters copied.
        """
        newcls = object.__new__(self.__class__)

        for key, val in self.__dict__.items():
            if key in self.param_names:  # copy only the parameter values
                newcls.__dict__[key] = copy.copy(val)
            else:
                newcls.__dict__[key] = val

        return newcls

    def deepcopy(self):
        """
        Return a deep copy of this model.

        Returns
        -------
        result : `GriddedPSFModel`
            A deep copy of this model.
        """
        return copy.deepcopy(self)

    @property
    def oversampling(self):
        """
        The integer oversampling factor(s) of the input ePSF images.

        If ``oversampling`` is a scalar then it will be used for both
        axes. If ``oversampling`` has two elements, they must be in
        ``(y, x)`` order.
        """
        return self._oversampling

    @oversampling.setter
    def oversampling(self, value):
        """
        Set the oversampling factor(s) of the input ePSF images.

        Parameters
        ----------
        value : int or tuple of int
            The integer oversampling factor(s) of the input ePSF images.
            If ``oversampling`` is a scalar then it will be used for both
            axes. If ``oversampling`` has two elements, they must be in
            ``(y, x)`` order.
        """
        self._oversampling = as_pair('oversampling', value, lower_bound=(0, 0))

    def _calc_bounding_box(self):
        """
        Set a bounding box defining the limits of the model.

        Returns
        -------
        bbox : tuple
            A bounding box defining the ((y_min, y_max), (x_min, x_max))
            limits of the model.
        """
        dy, dx = np.array(self.data.shape[1:]) / 2 / self.oversampling
        return ((self.y_0 - dy, self.y_0 + dy), (self.x_0 - dx, self.x_0 + dx))

    @property
    def bounding_box(self):
        """
        The bounding box of the model.

        Examples
        --------
        >>> from itertools import product
        >>> import numpy as np
        >>> from astropy.nddata import NDData
        >>> from photutils.psf import GaussianPSF, GriddedPSFModel
        >>> psfs = []
        >>> yy, xx = np.mgrid[0:101, 0:101]
        >>> for i in range(16):
        ...     theta = np.deg2rad(i * 10.0)
        ...     gmodel = GaussianPSF(flux=1, x_0=50, y_0=50, x_fwhm=10,
        ...                          y_fwhm=5, theta=theta)
        ...     psfs.append(gmodel(xx, yy))
        >>> xgrid = [0, 40, 160, 200]
        >>> ygrid = [0, 60, 140, 200]
        >>> meta = {}
        >>> meta['grid_xypos'] = list(product(xgrid, ygrid))
        >>> meta['oversampling'] = 4
        >>> nddata = NDData(psfs, meta=meta)
        >>> model = GriddedPSFModel(nddata, flux=1, x_0=0, y_0=0)
        >>> model.bounding_box  # doctest: +FLOAT_CMP
        ModelBoundingBox(
            intervals={
                x: Interval(lower=-12.625, upper=12.625)
                y: Interval(lower=-12.625, upper=12.625)
            }
            model=GriddedPSFModel(inputs=('x', 'y'))
            order='C'
        )
        """
        return self._calc_bounding_box()

    @lazyproperty
    def origin(self):
        """
        A 1D `~numpy.ndarray` (x, y) pixel coordinates within the
        model's 2D image of the origin of the coordinate system.
        """
        xyorigin = (np.array(self.data.shape) - 1) / 2
        return xyorigin[::-1]

    @lazyproperty
    def _interp_xyidx(self):
        """
        The x and y indices for the interpolator.
        """
        xidx = np.arange(self.data.shape[2])
        yidx = np.arange(self.data.shape[1])
        return xidx, yidx

    def _calc_interpolator(self, grid_idx):
        """
        Calculate the `~scipy.interpolate.RectBivariateSpline`
        interpolator for an input ePSF image at the given reference (x,
        y) position.

        The resulting interpolator is cached in the `_interpolator`
        dictionary for reuse.

        Parameters
        ----------
        grid_idx : int
            The index of the ePSF image in the reference grid.

        Returns
        -------
        interp : `~scipy.interpolate.RectBivariateSpline`
            The interpolator for the input ePSF image.
        """
        # check if the interpolator is already cached
        if grid_idx in self._interpolator:
            return self._interpolator[grid_idx]

        # RectBivariateSpline expects the data to be in (x, y) axis order
        data = self.data[grid_idx]
        interp = RectBivariateSpline(*self._interp_xyidx, data.T, kx=3, ky=3,
                                     s=0)

        # cache the interpolator for reuse
        self._interpolator[grid_idx] = interp

        return interp

    def _find_bounding_points(self, x, y):
        """
        Find the grid indices and reference (x, y) points of the four
        bounding grid points for a given (x, y) coordinate.

        If the point is outside the grid, the nearest grid points are
        selected. The input grid points do not need to be sorted.

        Parameters
        ----------
        x, y : float
            The (x_0, y_0) position of the model.

        Returns
        -------
        grid_idx : `~numpy.ndarray`
            The indices of the four bounding points in the sorted
            grid. The order is lower-left, lower-right, upper-left,
            upper-right.
        grid_xy : `~numpy.ndarray`
            The x and y coordinates of the four bounding points. The
            order is left, right, bottom, top.
        """
        # Find the insertion indices for x and y in the sorted grids
        xidx = np.searchsorted(self._xgrid, x) - 1
        yidx = np.searchsorted(self._ygrid, y) - 1

        # Clip the indices to valid ranges
        xidx = np.clip(xidx, 0, len(self._xgrid) - 2)
        yidx = np.clip(yidx, 0, len(self._ygrid) - 2)

        # Find the four bounding points in the sorted grid
        # (x0, y0) is the lower-left corner of the grid
        # (x1, y1) is the upper-right corner of the grid
        x0, x1 = self._xgrid[xidx], self._xgrid[xidx + 1]
        y0, y1 = self._ygrid[yidx], self._ygrid[yidx + 1]

        # Find the indices of these points in grid_xypos
        xcoords, ycoords = self.grid_xypos.T
        lower_left = np.where((xcoords == x0) & (ycoords == y0))[0][0]
        lower_right = np.where((xcoords == x1) & (ycoords == y0))[0][0]
        upper_left = np.where((xcoords == x0) & (ycoords == y1))[0][0]
        upper_right = np.where((xcoords == x1) & (ycoords == y1))[0][0]

        grid_idx = np.array((lower_left, lower_right, upper_left, upper_right))
        grid_xy = np.array((x0, x1, y0, y1))

        return grid_idx, grid_xy

    def _calc_bilinear_weights(self, xi, yi, grid_xy):
        """
        Calculate the bilinear interpolation weights for a given (xi,
        yi) coordinate and the four bounding grid points.

        Parameters
        ----------
        xi, yi : float
            The (x_0, y_0) position of the model.

        grid_xy : `~numpy.ndarray`
            The x and y coordinates of the four bounding points. The
            order is left, right, bottom, top.

        Returns
        -------
        weights : `~numpy.ndarray`
            The bilinear interpolation weights for the four bounding
            points. The order is lower-left, lower-right, upper-left,
            upper-right.
        """
        x0, x1, y0, y1 = grid_xy

        xi = np.clip(xi, x0, x1)
        yi = np.clip(yi, y0, y1)

        norm = (x1 - x0) * (y1 - y0)
        # lower-left, lower-right, upper-left, upper-right
        return np.array([(x1 - xi) * (y1 - yi), (xi - x0) * (y1 - yi),
                         (x1 - xi) * (yi - y0), (xi - x0) * (yi - y0)]) / norm

    def _calc_model_values(self, x_0, y_0, xi, yi):
        """
        Calculate the ePSF model at a given (x_0, y_0) model coordinate
        and the input (xi, yi) coordinate.

        Parameters
        ----------
        x_0, y_0 : float
            The (x, y) position of the model.

        xi, yi : float
            The input (x, y) coordinates at which the model is
            evaluated.

        Returns
        -------
        result : float or `~numpy.ndarray`
            The interpolated ePSF model at the input (x_0, y_0)
            coordinate.
        """
        grid_idx, grid_xy = self._find_bounding_points(x_0, y_0)
        interpolators = np.array([self._calc_interpolator(gidx)
                                  for gidx in grid_idx])
        weights = self._calc_bilinear_weights(x_0, y_0, grid_xy)

        idx = np.where(weights != 0)
        interpolators = interpolators[idx]
        weights = weights[idx]

        result = 0
        for interp, weight in zip(interpolators, weights, strict=True):
            result += interp(xi, yi, grid=False) * weight

        return result

    def evaluate(self, x, y, flux, x_0, y_0):
        """
        Calculate the ePSF model at the input coordinates for the given
        model parameters.

        Parameters
        ----------
        x, y : float or `~numpy.ndarray`
            The x and y positions at which to evaluate the model.

        flux : float
            The flux scaling factor for the model.

        x_0, y_0 : float
            The (x, y) position of the model.

        Returns
        -------
        evaluated_model : `~numpy.ndarray`
            The evaluated model.
        """
        if x.ndim > 2:
            msg = 'x and y must be 1D or 2D'
            raise ValueError(msg)

        # the base Model.__call__() method converts scalar inputs to
        # size-1 arrays before calling evaluate(), but we need scalar
        # values for the interpolator
        if not np.isscalar(x_0):
            x_0 = x_0[0]
        if not np.isscalar(y_0):
            y_0 = y_0[0]

        # now evaluate the ePSF at the (x_0, y_0) subpixel position on
        # the input (x, y) values
        xi = self.oversampling[1] * (np.asarray(x, dtype=float) - x_0)
        yi = self.oversampling[0] * (np.asarray(y, dtype=float) - y_0)
        xi += self.origin[0]
        yi += self.origin[1]

        if self.data.shape[0] == 1:
            # if there is only one ePSF, we do not need to perform
            # the bilinear interpolation
            evaluated_model = flux * self._calc_interpolator(0)(xi, yi,
                                                                grid=False)
        else:
            evaluated_model = flux * self._calc_model_values(x_0, y_0, xi, yi)

        if self.fill_value is not None:
            # set pixels that are outside the input pixel grid to the
            # fill_value to avoid extrapolation; these bounds match the
            # RegularGridInterpolator bounds
            ny, nx = self.data.shape[1:]
            invalid = (xi < 0) | (xi > nx - 1) | (yi < 0) | (yi > ny - 1)
            evaluated_model[invalid] = self.fill_value

        return evaluated_model

    @_plot_grid_docstring
    def plot_grid(self, *, ax=None, vmax_scale=None, peak_norm=False,
                  deltas=False, cmap='viridis', dividers=True,
                  divider_color='darkgray', divider_ls='-', figsize=None):
        plotter = _ModelGridPlotter(self)
        return plotter.plot_grid(ax=ax, vmax_scale=vmax_scale,
                                 peak_norm=peak_norm, deltas=deltas,
                                 cmap=cmap, dividers=dividers,
                                 divider_color=divider_color,
                                 divider_ls=divider_ls, figsize=figsize)


class STDPSFGrid:
    """
    Class to read and plot "STDPSF" format ePSF model grids.

    STDPSF files are FITS files that contain a 3D array of ePSFs with
    the header detailing where the fiducial ePSFs are located in the
    detector coordinate frame.

    The oversampling factor for STDPSF FITS files is assumed to be 4.

    Parameters
    ----------
    filename : str
        The name of the STDPDF FITS file. A URL can also be used.

    Examples
    --------
    >>> from photutils.psf import STDPSFGrid
    >>> psfgrid = STDPSFGrid('STDPSF_ACSWFC_F814W.fits')
    >>> fig = psfgrid.plot_grid()
    """

    def __init__(self, filename):
        grid_data = _read_stdpsf(filename)
        self.data = grid_data['data']
        self._xgrid = grid_data['xgrid']
        self._ygrid = grid_data['ygrid']
        xy_grid = [yx[::-1] for yx in itertools.product(self._ygrid,
                                                        self._xgrid)]
        oversampling = 4  # assumption for STDPSF files
        self.grid_xypos = xy_grid
        self.oversampling = as_pair('oversampling', oversampling,
                                    lower_bound=(0, 0))
        meta = {'grid_shape': (len(self._ygrid), len(self._xgrid)),
                'grid_xypos': xy_grid,
                'oversampling': oversampling}

        # try to get additional metadata from the filename because this
        # information is not currently available in the FITS headers
        file_meta = _get_metadata(filename, None)
        if file_meta is not None:
            meta.update(file_meta)

        self.meta = meta

    @_plot_grid_docstring
    def plot_grid(self, *, ax=None, vmax_scale=None, peak_norm=False,
                  deltas=False, cmap='viridis', dividers=True,
                  divider_color='darkgray', divider_ls='-', figsize=None):
        plotter = _ModelGridPlotter(self)
        return plotter.plot_grid(ax=ax, vmax_scale=vmax_scale,
                                 peak_norm=peak_norm, deltas=deltas,
                                 cmap=cmap, dividers=dividers,
                                 divider_color=divider_color,
                                 divider_ls=divider_ls, figsize=figsize)

    def __str__(self):
        cls_name = f'<{self.__class__.__module__}.{self.__class__.__name__}>'
        cls_info = []

        keys = ('STDPSF', 'detector', 'filter', 'grid_shape')
        for key in keys:
            if key in self.meta:
                name = key.capitalize() if key != 'STDPSF' else key
                cls_info.append((name, self.meta[key]))

        cls_info.extend([('Number of PSFs', len(self.grid_xypos)),
                         ('PSF shape (oversampled pixels)',
                          self.data.shape[1:]),
                         ('Oversampling', self.oversampling)])

        with np.printoptions(threshold=25, edgeitems=5):
            fmt = [f'{key}: {val}' for key, val in cls_info]

        return f'{cls_name}\n' + '\n'.join(fmt)

    def __repr__(self):
        return self.__str__()


with registry.delay_doc_updates(GriddedPSFModel):
    registry.register_reader('stdpsf', GriddedPSFModel, stdpsf_reader)
    registry.register_identifier('stdpsf', GriddedPSFModel, is_stdpsf)
    registry.register_reader('webbpsf', GriddedPSFModel, webbpsf_reader)
    registry.register_identifier('webbpsf', GriddedPSFModel, is_webbpsf)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the deblend module.
"""

from unittest.mock import patch

import numpy as np
import pytest
from astropy.modeling.models import Gaussian2D
from astropy.utils.exceptions import AstropyUserWarning
from numpy.testing import assert_allclose, assert_equal

from photutils.segmentation import (SegmentationImage, deblend_sources,
                                    detect_sources)
from photutils.segmentation.deblend import (_DeblendParams,
                                            _SingleSourceDeblender)
from photutils.utils._optional_deps import HAS_SKIMAGE


@pytest.mark.skipif(not HAS_SKIMAGE, reason='skimage is required')
class TestDeblendSources:
    @pytest.fixture(autouse=True)
    def setup(self):
        g1 = Gaussian2D(100, 50, 50, 5, 5)
        g2 = Gaussian2D(100, 35, 50, 5, 5)
        g3 = Gaussian2D(30, 70, 50, 5, 5)
        y, x = np.mgrid[0:100, 0:100]
        self.x = x
        self.y = y
        self.data = g1(x, y) + g2(x, y)
        self.data3 = self.data + g3(x, y)
        self.threshold = 10
        self.n_pixels = 5
        self.segm = detect_sources(self.data, self.threshold, self.n_pixels)
        self.segm3 = detect_sources(self.data3, self.threshold, self.n_pixels)

    @pytest.mark.parametrize('mode', ['exponential', 'linear', 'sinh'])
    def test_deblend_sources(self, mode):
        """
        Test deblend sources.
        """
        result = deblend_sources(self.data, self.segm, self.n_pixels,
                                 mode=mode, progress_bar=False)
        assert result.data.dtype == self.segm.data.dtype

        if mode == 'linear':
            # Test multiprocessing
            result2 = deblend_sources(self.data, self.segm,
                                      self.n_pixels,
                                      mode=mode,
                                      progress_bar=False,
                                      n_processes=2)
            assert_equal(result.data, result2.data)
            assert result2.data.dtype == self.segm.data.dtype

        assert result.n_labels == 2
        assert result.n_labels == len(result.slices)
        mask1 = (result.data == 1)
        mask2 = (result.data == 2)
        assert_allclose(len(result.data[mask1]), len(result.data[mask2]))
        assert_allclose(np.sum(self.data[mask1]), np.sum(self.data[mask2]))
        assert_allclose(np.nonzero(self.segm), np.nonzero(result))
        assert_equal(result.parent_to_deblended_labels, {1: [1, 2]})

    def test_deblend_multiple_sources(self):
        """
        Test deblend multiple sources.
        """
        g4 = Gaussian2D(100, 50, 15, 5, 5)
        g5 = Gaussian2D(100, 35, 15, 5, 5)
        g6 = Gaussian2D(100, 50, 85, 5, 5)
        g7 = Gaussian2D(100, 35, 85, 5, 5)
        x = self.x
        y = self.y
        data = self.data + g4(x, y) + g5(x, y) + g6(x, y) + g7(x, y)
        segm = detect_sources(data, self.threshold, self.n_pixels)
        result = deblend_sources(data, segm, self.n_pixels, progress_bar=False)
        assert result.n_labels == 6
        assert result.n_labels == len(result.slices)
        assert result.areas[0] == result.areas[1]
        assert result.areas[0] == result.areas[2]
        assert result.areas[0] == result.areas[3]
        assert result.areas[0] == result.areas[4]
        assert result.areas[0] == result.areas[5]

    def test_deblend_multiple_sources_with_neighbor(self):
        """
        Test deblend multiple sources with neighbor.
        """
        g1 = Gaussian2D(100, 50, 50, 20, 5, theta=45)
        g2 = Gaussian2D(100, 35, 50, 5, 5)
        g3 = Gaussian2D(100, 60, 20, 5, 5)

        x = self.x
        y = self.y
        data = (g1 + g2 + g3)(x, y)
        segm = detect_sources(data, self.threshold, self.n_pixels)
        result = deblend_sources(data, segm, self.n_pixels, progress_bar=False)
        assert result.n_labels == 3

    def test_deblend_labels(self):
        """
        Test deblend labels.
        """
        g1 = Gaussian2D(100, 50, 50, 20, 5, theta=45)
        g2 = Gaussian2D(100, 35, 50, 5, 5)
        g3 = Gaussian2D(100, 60, 20, 5, 5)
        x = self.x
        y = self.y
        data = (g1 + g2 + g3)(x, y)
        segm = detect_sources(data, self.threshold, self.n_pixels)
        result = deblend_sources(data, segm, self.n_pixels, labels=1,
                                 progress_bar=False)
        assert result.n_labels == 2

    @pytest.mark.parametrize(('contrast', 'nlabels'),
                             [(0.001, 6), (0.017, 5), (0.06, 4), (0.1, 3),
                              (0.15, 2), (0.45, 1)])
    def test_deblend_contrast(self, contrast, nlabels):
        """
        Test deblend contrast.
        """
        y, x = np.mgrid[0:51, 0:151]
        y0 = 25
        data = (Gaussian2D(9.5, 16, y0, 5, 5)(x, y)
                + Gaussian2D(51, 30, y0, 3, 3)(x, y)
                + Gaussian2D(30, 42, y0, 5, 5)(x, y)
                + Gaussian2D(80, 66, y0, 8, 8)(x, y)
                + Gaussian2D(71, 88, y0, 8, 8)(x, y)
                + Gaussian2D(18, 119, y0, 7, 7)(x, y))

        n_pixels = 5
        segm = detect_sources(data, 1.0, n_pixels)
        segm2 = deblend_sources(data, segm, n_pixels, mode='linear',
                                n_levels=32, contrast=contrast,
                                progress_bar=False)
        assert segm2.n_labels == nlabels

    def test_deblend_contrast_levels(self):
        """
        Test deblend contrast levels.

        Regression test for case where contrast=1.0.
        """
        y, x = np.mgrid[0:51, 0:151]
        y0 = 25
        data = (Gaussian2D(9.5, 16, y0, 5, 5)(x, y)
                + Gaussian2D(51, 30, y0, 3, 3)(x, y)
                + Gaussian2D(30, 42, y0, 5, 5)(x, y)
                + Gaussian2D(80, 66, y0, 8, 8)(x, y)
                + Gaussian2D(71, 88, y0, 8, 8)(x, y)
                + Gaussian2D(18, 119, y0, 7, 7)(x, y))

        n_pixels = 5
        segm = detect_sources(data, 1.0, n_pixels)
        for contrast in np.arange(1, 11) / 10.0:
            segm3 = deblend_sources(data, segm, n_pixels, mode='linear',
                                    n_levels=32, contrast=contrast,
                                    progress_bar=False)
            assert segm3.n_labels >= 1

    def test_deblend_connectivity(self):
        """
        Test deblend connectivity.
        """
        data = np.zeros((51, 51))
        data[15:36, 15:36] = 10.0
        data[14, 36] = 1.0
        data[13, 37] = 10
        data[14, 14] = 5.0
        data[13, 13] = 10.0
        data[36, 14] = 10.0
        data[37, 13] = 10.0
        data[36, 36] = 10.0
        data[37, 37] = 10.0

        segm = detect_sources(data, 0.1, 1, connectivity=4)
        assert segm.n_labels == 9
        segm2 = deblend_sources(data, segm, 1, mode='linear', connectivity=4,
                                progress_bar=False)
        assert segm2.n_labels == 9

        segm = detect_sources(data, 0.1, 1, connectivity=8)
        assert segm.n_labels == 1
        segm2 = deblend_sources(data, segm, 1, mode='linear', connectivity=8,
                                progress_bar=False)
        assert segm2.n_labels == 3

        match = 'Deblending failed for source'
        with pytest.raises(ValueError, match=match):
            deblend_sources(data, segm, 1, mode='linear', connectivity=4,
                            progress_bar=False)

    def test_deblend_label_assignment(self):
        """
        Test to ensure newly-deblended labels are unique.
        """
        y, x = np.mgrid[0:201, 0:101]
        y0a = 35
        y1a = 60
        yshift = 100
        y0b = y0a + yshift
        y1b = y1a + yshift
        data = (Gaussian2D(80, 36, y0a, 8, 8)(x, y)
                + Gaussian2D(71, 58, y1a, 8, 8)(x, y)
                + Gaussian2D(30, 36, y1a, 7, 7)(x, y)
                + Gaussian2D(30, 58, y0a, 7, 7)(x, y)
                + Gaussian2D(80, 36, y0b, 8, 8)(x, y)
                + Gaussian2D(71, 58, y1b, 8, 8)(x, y)
                + Gaussian2D(30, 36, y1b, 7, 7)(x, y)
                + Gaussian2D(30, 58, y0b, 7, 7)(x, y))

        n_pixels = 5
        segm1 = detect_sources(data, 5.0, n_pixels)
        segm2 = deblend_sources(data, segm1, n_pixels, mode='linear',
                                n_levels=32, contrast=0.3, progress_bar=False)
        assert segm2.n_labels == 4

    @pytest.mark.parametrize('mode', ['exponential', 'linear'])
    def test_deblend_sources_norelabel(self, mode):
        """
        Test deblend sources norelabel.
        """
        result = deblend_sources(self.data, self.segm, self.n_pixels,
                                 mode=mode, relabel=False, progress_bar=False)
        assert result.n_labels == 2
        assert_equal(result.labels, [2, 3])
        assert_equal(result.parent_to_deblended_labels, {1: [2, 3]})
        assert len(result.slices) <= result.max_label
        assert len(result.slices) == result.n_labels
        assert_allclose(np.nonzero(self.segm), np.nonzero(result))

    @pytest.mark.parametrize('mode', ['exponential', 'linear'])
    def test_deblend_three_sources(self, mode):
        """
        Test deblend three sources.
        """
        result = deblend_sources(self.data3, self.segm3, self.n_pixels,
                                 mode=mode, progress_bar=False)
        assert result.n_labels == 3
        assert_allclose(np.nonzero(self.segm3), np.nonzero(result))

    def test_segmentation_image(self):
        """
        Test segmentation image.
        """
        segm_wrong = np.ones((2, 2), dtype=int)  # ndarray
        match = 'segmentation_image must be a SegmentationImage'
        with pytest.raises(TypeError, match=match):
            deblend_sources(self.data, segm_wrong, self.n_pixels,
                            progress_bar=False)

        segm_wrong = SegmentationImage(segm_wrong)  # wrong shape
        match = 'segmentation_image must have the same shape as data'
        with pytest.raises(ValueError, match=match):
            deblend_sources(self.data, segm_wrong, self.n_pixels,
                            progress_bar=False)

    def test_invalid_n_levels(self):
        """
        Test invalid n_levels.
        """
        match = 'n_levels must be >= 1'
        with pytest.raises(ValueError, match=match):
            deblend_sources(self.data, self.segm, self.n_pixels, n_levels=0,
                            progress_bar=False)

    def test_invalid_contrast(self):
        """
        Test invalid contrast.
        """
        match = 'contrast must be >= 0 and <= 1'
        with pytest.raises(ValueError, match=match):
            deblend_sources(self.data, self.segm, self.n_pixels, contrast=-1,
                            progress_bar=False)

    def test_invalid_mode(self):
        """
        Test invalid mode.
        """
        match = "mode must be 'exponential', 'linear', or 'sinh'"
        with pytest.raises(ValueError, match=match):
            deblend_sources(self.data, self.segm, self.n_pixels,
                            mode='invalid', progress_bar=False)

    def test_invalid_connectivity(self):
        """
        Test invalid connectivity.
        """
        match = 'Invalid connectivity'
        with pytest.raises(ValueError, match=match):
            deblend_sources(self.data, self.segm, self.n_pixels,
                            connectivity='invalid', progress_bar=False)

    def test_constant_source(self):
        """
        Test constant source.
        """
        data = self.data.copy()
        data[data.nonzero()] = 1.0
        result = deblend_sources(data, self.segm, self.n_pixels,
                                 progress_bar=False)
        assert_allclose(result, self.segm)

    def test_source_with_negval(self):
        """
        Test source with negval.
        """
        data = self.data.copy()
        data -= 20
        match = 'The deblending mode of one or more source labels from the'
        with pytest.warns(AstropyUserWarning, match=match):
            segm = deblend_sources(data, self.segm, self.n_pixels,
                                   progress_bar=False)
        assert segm.info['warnings']['nonposmin']['input_labels'] == 1

    def test_source_zero_min(self):
        """
        Test source zero min.
        """
        data = self.data.copy()
        data -= data[self.segm.data > 0].min()
        match = 'The deblending mode of one or more source labels from the'
        with pytest.warns(AstropyUserWarning, match=match):
            segm = deblend_sources(data, self.segm, self.n_pixels,
                                   progress_bar=False)
        assert segm.info['warnings']['nonposmin']['input_labels'] == 1

    def test_connectivity(self):
        """
        Test connectivity.

        Regression test for #341.
        """
        data = np.zeros((3, 3))
        data[0, 0] = 2
        data[1, 1] = 2
        data[2, 2] = 1
        segm = np.zeros(data.shape, dtype=int)
        segm[data.nonzero()] = 1
        segm = SegmentationImage(segm)
        data = data * 100.0
        segm_deblend = deblend_sources(data, segm, n_pixels=1, connectivity=8,
                                       progress_bar=False)
        assert segm_deblend.n_labels == 1
        match = 'Deblending failed for source'
        with pytest.raises(ValueError, match=match):
            deblend_sources(data, segm, n_pixels=1, connectivity=4,
                            progress_bar=False)

    def test_data_nan(self):
        """
        Test that deblending occurs even if the data within a segment
        contains one or more NaNs.

        Regression test for #658.
        """
        data = self.data.copy()
        data[50, 50] = np.nan
        segm2 = deblend_sources(data, self.segm, 5, progress_bar=False)
        assert segm2.n_labels == 2

    def test_watershed(self):
        """
        Test that the watershed input mask is a bool array.

        With scikit-image >= 0.13, the mask must be a bool array. In
        particular, if the mask array contains label 512, the watershed
        algorithm fails.
        """
        segm = self.segm.copy()
        segm.reassign_label(1, 512)
        result = deblend_sources(self.data, segm, self.n_pixels,
                                 progress_bar=False)
        assert result.n_labels == 2

    def test_nondetection(self):
        """
        Test for case where no sources are detected at one of the
        threshold levels.

        For this case, a `NoDetectionsWarning` should not be raised when
        deblending sources.
        """
        data = np.copy(self.data3)
        data[50, 50] = 1000.0
        data[50, 70] = 500.0
        self.segm = detect_sources(data, self.threshold, self.n_pixels)
        deblend_sources(data, self.segm, self.n_pixels, progress_bar=False)

    def test_nonconsecutive_labels(self):
        """
        Test nonconsecutive labels.
        """
        segm = self.segm.copy()
        segm.reassign_label(1, 1000)
        result = deblend_sources(self.data, segm, self.n_pixels,
                                 progress_bar=False)
        assert result.n_labels == 2

    def test_single_source_methods(self):
        """
        Test the multithreshold and make_markers methods of the
        _SingleSourceDeblender class.

        These methods are useful for debugging but are not currently
        used by the deblend_sources function.
        """
        data = self.data3
        segm = self.segm3
        n_pixels = 5
        footprint = np.ones((3, 3))
        deblend_params = _DeblendParams(n_pixels, footprint, 32, 0.001,
                                        'linear')
        single_debl = _SingleSourceDeblender(data, segm.data, 1,
                                             deblend_params)
        segms = single_debl.multithreshold()
        assert len(segms) == 32

        markers = single_debl.make_markers(return_all=True)
        assert len(markers) == 19

    def test_deblend_progress_bar(self):
        """
        Test deblend_sources with progress_bar=True (serial).
        """
        result = deblend_sources(self.data, self.segm, self.n_pixels,
                                 mode='linear', progress_bar=True)
        assert result.n_labels == 2

    def test_deblend_nproc_none(self):
        """
        Test deblend_sources with n_processes=None (auto-detect CPU count).
        """
        result = deblend_sources(self.data, self.segm, self.n_pixels,
                                 mode='linear', progress_bar=False,
                                 n_processes=None)
        assert result.n_labels == 2


@pytest.mark.skipif(not HAS_SKIMAGE, reason='skimage is required')
def test_nmarkers_fallback():
    """
    Test that if there are too many markers, a warning is raised.
    """
    size = 51
    data1 = np.resize([0, 0, 1, 1], size)
    data1 = np.abs(data1 - np.atleast_2d(data1).T) + 2

    for i in range(size):
        if i % 2 == 0:
            data1[i, :] = 1
            data1[:, i] = 1

    data = np.zeros((101, 101))
    data[25:25 + size, 25:25 + size] = data1
    data[50:60, 50:60] = 10.0

    segm = detect_sources(data, 0.01, 10)
    match = 'The deblending mode of one or more source labels from the'
    with pytest.warns(AstropyUserWarning, match=match):
        segm2 = deblend_sources(data, segm, 1, mode='exponential')
    assert segm2.info['warnings']['nmarkers']['input_labels'][0] == 1
    mesg = segm2.info['warnings']['nmarkers']['message']
    assert mesg.startswith('Deblending mode changed')


@pytest.mark.skipif(not HAS_SKIMAGE, reason='skimage is required')
def test_nmarkers_fallback_multiproc():
    """
    Test the nmarkers fallback warning via multiprocessing (n_processes=2).
    This covers the multiprocessing result-processing block for nmarkers.
    """
    size = 51
    data1 = np.resize([0, 0, 1, 1], size)
    data1 = np.abs(data1 - np.atleast_2d(data1).T) + 2

    for i in range(size):
        if i % 2 == 0:
            data1[i, :] = 1
            data1[:, i] = 1

    data = np.zeros((101, 101))
    data[25:25 + size, 25:25 + size] = data1
    data[50:60, 50:60] = 10.0

    segm = detect_sources(data, 0.01, 10)
    match = 'The deblending mode of one or more source labels from the'
    with pytest.warns(AstropyUserWarning, match=match):
        segm2 = deblend_sources(data, segm, 1, mode='exponential',
                                n_processes=2)
    assert segm2.info['warnings']['nmarkers']['input_labels'][0] == 1


@pytest.mark.skipif(not HAS_SKIMAGE, reason='skimage is required')
def test_nonposmin_multiproc():
    """
    Test nonposmin warning via multiprocessing (n_processes=2).
    This covers the multiprocessing result-processing block for
    nonposmin.
    """
    g1 = Gaussian2D(100, 50, 50, 8, 8)
    g2 = Gaussian2D(100, 35, 50, 8, 8)
    yy, xx = np.mgrid[0:101, 0:101]
    data = g1(xx, yy) + g2(xx, yy) - 20  # negative values

    segm = detect_sources(data + 20, 10, 5)  # detect sources on positive data
    match = 'The deblending mode of one or more source labels from the'
    with pytest.warns(AstropyUserWarning, match=match):
        segm2 = deblend_sources(data, segm, 5, progress_bar=False,
                                n_processes=2)
    assert 'nonposmin' in segm2.info['warnings']


@pytest.mark.skipif(not HAS_SKIMAGE, reason='skimage is required')
def test_nmarkers_fallback_returns_none():
    """
    Test that deblend_source returns None when make_markers returns
    None on the linear-mode fallback (second attempt after >200
    markers).
    """
    # Create a source with varying data values so source_min != source_max
    data = np.ones((20, 20)) * 10.0
    data[5:15, 5:15] = 50.0
    data[8:12, 8:12] = 100.0  # peak in center
    segment = np.zeros((20, 20), dtype=int)
    segment[5:15, 5:15] = 1

    deblend_params = _DeblendParams(5, np.ones((3, 3)), 32, 0.001,
                                    'exponential')

    deblender = _SingleSourceDeblender(data, segment, 1, deblend_params)

    call_count = [0]

    def mock_make_markers(*, _return_all=False):
        call_count[0] += 1
        if call_count[0] == 1:
            # First call: return markers with > 200 labels
            markers = np.zeros((20, 20), dtype=int)
            for i in range(201):
                r, c = divmod(i, 20)
                if r < 20 and c < 20:
                    markers[r, c] = i + 1
            return markers
        # Second call (linear fallback): return None
        return None

    with patch.object(deblender, 'make_markers', mock_make_markers):
        result = deblender.deblend_source()

    assert result is None



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for making example WCS objects.
"""

import astropy.units as u
import numpy as np
from astropy import coordinates as coord
from astropy.modeling import models
from astropy.wcs import WCS

from photutils.utils._deprecation import deprecated_positional_kwargs

__all__ = ['make_gwcs', 'make_wcs']

__doctest_requires__ = {'make_gwcs': ['gwcs']}


@deprecated_positional_kwargs(since='3.0', until='4.0')
def make_wcs(shape, galactic=False):
    """
    Create a simple celestial `~astropy.wcs.WCS` object in either the
    ICRS or Galactic coordinate frame.

    Parameters
    ----------
    shape : 2-tuple of int
        The shape of the 2D array to be used with the output
        `~astropy.wcs.WCS` object.

    galactic : bool, optional
        If `True`, then the output WCS will be in the Galactic
        coordinate frame. If `False` (default), then the output WCS will
        be in the ICRS coordinate frame.

    Returns
    -------
    wcs : `astropy.wcs.WCS` object
        The world coordinate system (WCS) transformation.

    See Also
    --------
    make_gwcs

    Notes
    -----
    The `make_gwcs` function returns an equivalent WCS transformation to
    this one, but as a `gwcs.wcs.WCS` object.

    Examples
    --------
    >>> from photutils.datasets import make_wcs
    >>> shape = (100, 100)
    >>> wcs = make_wcs(shape)
    >>> print(wcs.wcs.crpix)  # doctest: +FLOAT_CMP
    [50. 50.]
    >>> print(wcs.wcs.crval)  # doctest: +FLOAT_CMP
    [197.8925      -1.36555556]
    >>> skycoord = wcs.pixel_to_world(42, 57)
    >>> print(skycoord)  # doctest: +FLOAT_CMP
    <SkyCoord (ICRS): (ra, dec) in deg
        (197.89278975, -1.36561284)>
    """
    wcs = WCS(naxis=2)
    rho = np.pi / 3.0
    scale = 0.1 / 3600.0  # 0.1 arcsec/pixel in deg/pix

    wcs.pixel_shape = shape
    wcs.wcs.crpix = [shape[1] / 2, shape[0] / 2]  # 1-indexed (x, y)
    wcs.wcs.crval = [197.8925, -1.36555556]
    wcs.wcs.cunit = ['deg', 'deg']
    wcs.wcs.cd = [[-scale * np.cos(rho), scale * np.sin(rho)],
                  [scale * np.sin(rho), scale * np.cos(rho)]]
    if not galactic:
        wcs.wcs.radesys = 'ICRS'
        wcs.wcs.ctype = ['RA---TAN', 'DEC--TAN']
    else:
        wcs.wcs.ctype = ['GLON-CAR', 'GLAT-CAR']

    return wcs


@deprecated_positional_kwargs(since='3.0', until='4.0')
def make_gwcs(shape, galactic=False):
    """
    Create a simple celestial gWCS object in either the ICRS or Galactic
    coordinate frame.

    This function requires the `gwcs
    <https://github.com/spacetelescope/gwcs>`_ package.

    Parameters
    ----------
    shape : 2-tuple of int
        The shape of the 2D array to be used with the output
        `~gwcs.wcs.WCS` object.

    galactic : bool, optional
        If `True`, then the output WCS will be in the Galactic
        coordinate frame. If `False` (default), then the output WCS will
        be in the ICRS coordinate frame.

    Returns
    -------
    wcs : `gwcs.wcs.WCS` object
        The generalized world coordinate system (WCS) transformation.

    See Also
    --------
    make_wcs

    Notes
    -----
    The `make_wcs` function returns an equivalent WCS transformation to
    this one, but as an `astropy.wcs.WCS` object.

    Examples
    --------
    >>> from photutils.datasets import make_gwcs
    >>> shape = (100, 100)
    >>> gwcs = make_gwcs(shape)
    >>> print(gwcs)
      From      Transform
    -------- ----------------
    detector linear_transform
        icrs             None
    >>> skycoord = gwcs.pixel_to_world(42, 57)
    >>> print(skycoord)  # doctest: +FLOAT_CMP
    <SkyCoord (ICRS): (ra, dec) in deg
        (197.89278975, -1.36561284)>
    """
    from gwcs import coordinate_frames as cf
    from gwcs import wcs as gwcs_wcs

    rho = np.pi / 3.0
    scale = 0.1 / 3600.0  # 0.1 arcsec/pixel in deg/pix

    shift_by_crpix = (models.Shift((-shape[1] / 2) + 1)
                      & models.Shift((-shape[0] / 2) + 1))

    cd_matrix = np.array([[-scale * np.cos(rho), scale * np.sin(rho)],
                          [scale * np.sin(rho), scale * np.cos(rho)]])

    rotation = models.AffineTransformation2D(cd_matrix, translation=[0, 0])
    rotation.inverse = models.AffineTransformation2D(
        np.linalg.inv(cd_matrix), translation=[0, 0])

    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(197.8925, -1.36555556,
                                                       180.0)

    det2sky = shift_by_crpix | rotation | tan | celestial_rotation
    det2sky.name = 'linear_transform'

    detector_frame = cf.Frame2D(name='detector', axes_names=('x', 'y'),
                                unit=(u.pix, u.pix))

    if galactic:
        sky_frame = cf.CelestialFrame(reference_frame=coord.Galactic(),
                                      name='galactic', unit=(u.deg, u.deg))
    else:
        sky_frame = cf.CelestialFrame(reference_frame=coord.ICRS(),
                                      name='icrs', unit=(u.deg, u.deg))

    pipeline = [(detector_frame, det2sky), (sky_frame, None)]

    return gwcs_wcs.WCS(pipeline)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for deprecation warnings when optional arguments are passed
positionally.
"""

import numpy as np
import pytest
from astropy.utils.exceptions import AstropyDeprecationWarning

from photutils.background import (BiweightLocationBackground,
                                  BiweightScaleBackgroundRMS, LocalBackground,
                                  MADStdBackgroundRMS, MeanBackground,
                                  MedianBackground, ModeEstimatorBackground,
                                  SExtractorBackground, StdBackgroundRMS)

BKG_CLASSES = [MeanBackground,
               MedianBackground,
               ModeEstimatorBackground,
               SExtractorBackground,
               BiweightLocationBackground]
BKGRMS_CLASSES = [StdBackgroundRMS,
                  MADStdBackgroundRMS,
                  BiweightScaleBackgroundRMS]


class TestBackgroundBasePositionalKwargs:
    """
    Test that __call__ on BackgroundBase subclasses warns for positional
    optional args.
    """

    def setup_method(self):
        self.data = np.arange(100, dtype=float)

    @pytest.mark.parametrize('cls', BKG_CLASSES)
    def test_call_positional_warns(self, cls):
        bkg = cls()
        match = '__call__'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            bkg(self.data, None)

        match = 'calc_background'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            bkg.calc_background(self.data, None)

    @pytest.mark.parametrize('cls', BKG_CLASSES)
    def test_call_keyword_no_warning(self, cls):
        bkg = cls()
        bkg(self.data, axis=None)


class TestBackgroundRMSBasePositionalKwargs:
    """
    Test that __call__ on BackgroundRMSBase subclasses warns for
    positional optional args.
    """

    def setup_method(self):
        self.data = np.arange(100, dtype=float)

    @pytest.mark.parametrize('cls', BKGRMS_CLASSES)
    def test_call_positional_warns(self, cls):
        bkgrms = cls()
        match = '__call__'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            bkgrms(self.data, None)

        match = 'calc_background_rms'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            bkgrms.calc_background_rms(self.data, None)

    @pytest.mark.parametrize('cls', BKGRMS_CLASSES)
    def test_call_keyword_no_warning(self, cls):
        bkgrms = cls()
        bkgrms(self.data, axis=None)


class TestLocalBackgroundPositionalKwargs:
    """
    Test that LocalBackground warns for positional optional args.
    """

    def setup_method(self):
        self.data = np.ones((101, 101))

    def test_init_positional_warns(self):
        match = '__init__'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            LocalBackground(5, 10, MedianBackground())

    def test_init_keyword_no_warning(self):
        LocalBackground(5, 10, bkg_estimator=MedianBackground())

    def test_call_positional_warns(self):
        local_bkg = LocalBackground(5, 10)
        mask = np.zeros((101, 101), dtype=bool)
        match = '__call__'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            local_bkg(self.data, 50, 50, mask)

    def test_call_keyword_no_warning(self):
        local_bkg = LocalBackground(5, 10)
        mask = np.zeros((101, 101), dtype=bool)
        local_bkg(self.data, 50, 50, mask=mask)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the daofinder module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.utils.exceptions import AstropyDeprecationWarning
from numpy.testing import assert_array_equal

from photutils.detection.daofinder import DAOStarFinder
from photutils.utils.exceptions import NoDetectionsWarning


class TestDAOStarFinder:
    """
    Test the DAOStarFinder class.
    """

    def test_find(self, data):
        """
        Test basic source detection and unit handling.
        """
        units = u.Jy
        threshold = 5.0
        fwhm = 1.0
        finder0 = DAOStarFinder(threshold, fwhm)
        finder1 = DAOStarFinder(threshold * units, fwhm)

        tbl0 = finder0(data)
        tbl1 = finder1(data << units)
        assert_array_equal(tbl0, tbl1)

        assert np.min(tbl0['flux']) > 150

        # Test that sources are returned with threshold = 0
        finder = DAOStarFinder(0, fwhm)
        tbl = finder(data)
        assert len(tbl) == 25

    def test_inputs(self):
        """
        Test that invalid inputs raise appropriate errors.
        """
        match = 'fwhm must be a scalar value'
        with pytest.raises(TypeError, match=match):
            DAOStarFinder(threshold=3.0, fwhm=np.ones((2, 2)))

        match = 'fwhm must be positive'
        for fwhm in (-10, 0):
            with pytest.raises(ValueError, match=match):
                DAOStarFinder(threshold=3.0, fwhm=fwhm)

        match = 'ratio must be > 0 and <= 1.0'
        with pytest.raises(ValueError, match=match):
            DAOStarFinder(threshold=3.0, fwhm=2, ratio=-10)

        match = 'sigma_radius must be positive'
        with pytest.raises(ValueError, match=match):
            DAOStarFinder(threshold=3.0, fwhm=2, sigma_radius=-10)

        match = 'n_brightest must be > 0'
        with pytest.raises(ValueError, match=match):
            DAOStarFinder(threshold=10, fwhm=1.5, n_brightest=-1)

        match = 'n_brightest must be an integer'
        with pytest.raises(ValueError, match=match):
            DAOStarFinder(threshold=10, fwhm=1.5, n_brightest=3.1)

        xycoords = np.array([[1, 2, 3, 4], [5, 6, 7, 8]])
        match = 'xycoords must be shaped as an Nx2 array'
        with pytest.raises(ValueError, match=match):
            DAOStarFinder(threshold=10, fwhm=1.5, xycoords=xycoords)

    @pytest.mark.parametrize(('theta', 'expected'), [
        (400.0, 40.0),
        (-30.0, 330.0),
        (360.0, 0.0),
        (0.0, 0.0),
    ])
    def test_theta_normalization(self, theta, expected):
        """
        Test that theta values are normalized to [0, 360).
        """
        finder = DAOStarFinder(threshold=5.0, fwhm=3.0, theta=theta)
        assert finder.theta == expected

    def test_nosources(self, data):
        """
        Test that no sources returns None with a warning.
        """
        match = 'No sources were found'
        finder = DAOStarFinder(threshold=100, fwhm=2)
        with pytest.warns(NoDetectionsWarning, match=match):
            tbl = finder(data)
        assert tbl is None

        finder = DAOStarFinder(threshold=1, fwhm=2)
        with pytest.warns(NoDetectionsWarning, match=match):
            tbl = finder(-data)
        assert tbl is None

    def test_exclude_border(self):
        """
        Test that border sources are excluded.
        """
        data = np.zeros((9, 9))
        data[0, 0] = 1
        data[2, 2] = 1
        data[4, 4] = 1
        data[6, 6] = 1

        finder0 = DAOStarFinder(threshold=0.1, fwhm=0.5, exclude_border=False)
        finder1 = DAOStarFinder(threshold=0.1, fwhm=0.5, exclude_border=True)
        tbl0 = finder0(data)
        tbl1 = finder1(data)
        assert len(tbl0) > len(tbl1)

    def test_mask(self, data):
        """
        Test source detection with a mask.
        """
        finder = DAOStarFinder(threshold=1.0, fwhm=1.5)
        mask = np.zeros(data.shape, dtype=bool)
        mask[0:50, :] = True
        tbl0 = finder(data)
        tbl1 = finder(data, mask=mask)
        assert len(tbl0) > len(tbl1)

    def test_mask_int(self, data):
        """
        Test that an integer mask gives the same result as a boolean
        mask.
        """
        finder = DAOStarFinder(threshold=1.0, fwhm=1.5)
        bool_mask = np.zeros(data.shape, dtype=bool)
        bool_mask[0:50, :] = True
        int_mask = bool_mask.astype(int)

        tbl_bool = finder(data, mask=bool_mask)
        tbl_int = finder(data, mask=int_mask)
        assert_array_equal(tbl_bool, tbl_int)

    def test_xycoords(self, data):
        """
        Test source detection at specified coordinates.
        """
        finder0 = DAOStarFinder(threshold=8.0, fwhm=2)
        tbl0 = finder0(data)
        xycoords = list(zip(tbl0['x_centroid'],
                            tbl0['y_centroid'], strict=True))
        xycoords = np.round(xycoords).astype(int)

        finder1 = DAOStarFinder(threshold=8.0, fwhm=2, xycoords=xycoords)
        tbl1 = finder1(data)
        assert_array_equal(tbl0, tbl1)

    def test_min_separation(self, data):
        """
        Test the min_separation parameter.
        """
        threshold = 1.0
        fwhm = 1.0
        finder1 = DAOStarFinder(threshold, fwhm)
        tbl1 = finder1(data)
        assert finder1.min_separation == 2.5 * fwhm

        finder2 = DAOStarFinder(threshold, fwhm, min_separation=10.0)
        tbl2 = finder2(data)
        assert len(tbl1) > len(tbl2)

        match = 'min_separation must be >= 0'
        with pytest.raises(ValueError, match=match):
            DAOStarFinder(threshold=10, fwhm=1.5, min_separation=-1.0)

    def test_min_separation_default(self):
        """
        Test that the default min_separation (None) gives 2.5 * fwhm.
        """
        fwhm = 2.0
        finder = DAOStarFinder(threshold=1.0, fwhm=fwhm)
        assert finder.min_separation == 2.5 * fwhm

        finder_old = DAOStarFinder(threshold=1.0, fwhm=fwhm,
                                   min_separation=0)
        assert finder_old.min_separation == 0

    def test_brightest_filtering(self, data):
        """
        Test that only the top brightest sources are selected.
        """
        n_brightest = 10
        finder = DAOStarFinder(threshold=1.0, fwhm=1.5,
                               roundness_range=(-np.inf, np.inf),
                               sharpness_range=(-np.inf, np.inf),
                               n_brightest=n_brightest)
        tbl = finder(data)
        assert len(tbl) == n_brightest

        # Combined with peak_max
        peak_max = 8
        finder = DAOStarFinder(threshold=1.0, fwhm=1.5,
                               roundness_range=(-np.inf, np.inf),
                               sharpness_range=(-np.inf, np.inf),
                               n_brightest=n_brightest,
                               peak_max=peak_max)
        tbl = finder(data)
        assert len(tbl) == 5

    def test_sharpness(self, data):
        """
        Test that no sources pass the sharpness criteria.
        """
        finder = DAOStarFinder(threshold=1, fwhm=1.0,
                               sharpness_range=(1.0, 1.0))
        match = 'Sources were found, but none pass'
        with pytest.warns(NoDetectionsWarning, match=match):
            tbl = finder(data)
        assert tbl is None

    @pytest.mark.parametrize('sharpness_range', [0.5, (0.5,), (1, 2, 3)])
    def test_invalid_sharpness_range(self, sharpness_range):
        match = 'sharpness_range must be a 2-element .* tuple'
        with pytest.raises(ValueError, match=match):
            DAOStarFinder(threshold=1, fwhm=1.0,
                          sharpness_range=sharpness_range)

    def test_sharpness_range_none(self, data):
        """
        Test that sharpness_range=None disables sharpness filtering.
        """
        finder_none = DAOStarFinder(threshold=1, fwhm=1.0,
                                    roundness_range=None,
                                    sharpness_range=None)
        tbl_none = finder_none(data)
        assert tbl_none is not None

        finder_strict = DAOStarFinder(threshold=1, fwhm=1.0,
                                      roundness_range=None,
                                      sharpness_range=(1.0, 1.0))
        match = 'Sources were found, but none pass'
        with pytest.warns(NoDetectionsWarning, match=match):
            tbl_strict = finder_strict(data)
        assert tbl_strict is None
        assert len(tbl_none) >= 1

    def test_roundness(self, data):
        """
        Test that no sources pass the roundness criteria.
        """
        match = 'Sources were found, but none pass'
        finder = DAOStarFinder(threshold=1, fwhm=1.0,
                               roundness_range=(1.0, 1.0))
        with pytest.warns(NoDetectionsWarning, match=match):
            tbl = finder(data)
        assert tbl is None

    @pytest.mark.parametrize('roundness_range', [0.5, (0.5,), (1, 2, 3)])
    def test_invalid_roundness_range(self, roundness_range):
        match = 'roundness_range must be a 2-element .* tuple'
        with pytest.raises(ValueError, match=match):
            DAOStarFinder(threshold=1, fwhm=1.0,
                          roundness_range=roundness_range)

    def test_roundness_range_none(self, data):
        """
        Test that roundness_range=None disables roundness filtering.
        """
        finder_none = DAOStarFinder(threshold=1, fwhm=1.0,
                                    sharpness_range=None,
                                    roundness_range=None)
        tbl_none = finder_none(data)
        assert tbl_none is not None

        finder_strict = DAOStarFinder(threshold=1, fwhm=1.0,
                                      sharpness_range=None,
                                      roundness_range=(1.0, 1.0))
        match = 'Sources were found, but none pass'
        with pytest.warns(NoDetectionsWarning, match=match):
            tbl_strict = finder_strict(data)
        assert tbl_strict is None
        assert len(tbl_none) >= 1

    def test_peak_max(self, data):
        """
        Test that no sources pass the peak_max criteria.
        """
        match = 'Sources were found, but none pass'
        finder = DAOStarFinder(threshold=1, fwhm=1.0, peak_max=1.0)
        with pytest.warns(NoDetectionsWarning, match=match):
            tbl = finder(data)
        assert tbl is None

    def test_peak_max_filtering(self, data):
        """
        Test that sources with peak >= peak_max are filtered out.
        """
        peak_max = 8
        finder0 = DAOStarFinder(threshold=1.0, fwhm=1.5,
                                roundness_range=(-np.inf, np.inf),
                                sharpness_range=(-np.inf, np.inf))
        finder1 = DAOStarFinder(threshold=1.0, fwhm=1.5,
                                roundness_range=(-np.inf, np.inf),
                                sharpness_range=(-np.inf, np.inf),
                                peak_max=peak_max)

        tbl0 = finder0(data)
        tbl1 = finder1(data)
        assert len(tbl0) > len(tbl1)
        assert all(tbl1['peak'] <= peak_max)

    def test_single_detected_source(self, data):
        """
        Test detection and slicing with a single source.
        """
        finder = DAOStarFinder(7.9, 2, n_brightest=1)
        mask = np.zeros(data.shape, dtype=bool)
        mask[0:50] = True
        tbl = finder(data, mask=mask)
        assert len(tbl) == 1

        # Test slicing with scalar catalog to improve coverage
        cat = finder._get_raw_catalog(data, mask=mask)
        assert cat.isscalar
        flux = cat.flux[0]  # evaluate the flux so it can be sliced
        assert cat[0].flux == flux

    def test_interval_ends_included(self):
        """
        Test that filter interval endpoints are inclusive.
        """
        # https://github.com/astropy/photutils/issues/1977
        data = np.zeros((46, 64))

        x = 33
        y = 21

        data[y - 1: y + 2, x - 1: x + 2] = [
            [1.0, 2.0, 1.0],
            [2.0, 1.0e20, 2.0],
            [1.0, 2.0, 1.0],
        ]

        finder = DAOStarFinder(
            threshold=0,
            fwhm=2.5,
            roundness_range=(0, 1.0),
            sharpness_range=(0.2, 1.407913491884342),
            peak_max=1.0e20,
        )
        tbl = finder.find_stars(data)

        assert len(tbl) == 1
        assert tbl[0]['roundness1'] < 1.e-15
        assert tbl[0]['roundness2'] == 0.0
        assert tbl[0]['peak'] == 1.0e20

    def test_data_not_mutated(self, data):
        """
        Test that input data is not mutated by find_stars.
        """
        data_copy = data.copy()
        finder = DAOStarFinder(threshold=1.0, fwhm=1.5)
        finder(data)
        assert_array_equal(data, data_copy)

    def test_data_not_mutated_with_mask(self, data):
        """
        Test that input data is not mutated when a mask is used.
        """
        data_copy = data.copy()
        mask = np.zeros(data.shape, dtype=bool)
        mask[0:50] = True
        finder = DAOStarFinder(threshold=1.0, fwhm=1.5)
        finder(data, mask=mask)
        assert_array_equal(data, data_copy)

    def test_repr(self):
        """
        Test the __repr__ of DAOStarFinder.
        """
        finder = DAOStarFinder(threshold=5.0, fwhm=3.0)
        repr_ = repr(finder)
        assert 'DAOStarFinder(' in repr_
        assert 'threshold=5.0' in repr_
        assert 'fwhm=3.0' in repr_
        assert 'ratio=1.0' in repr_
        assert 'xycoords=None' in repr_

    def test_str(self):
        """
        Test the __str__ of DAOStarFinder.
        """
        finder = DAOStarFinder(threshold=5.0, fwhm=3.0)
        str_ = str(finder)
        assert 'DAOStarFinder' in str_
        assert 'threshold: 5.0' in str_
        assert 'fwhm: 3.0' in str_

    def test_repr_with_xycoords(self):
        """
        Test that __repr__ shows array shape when xycoords are provided.
        """
        xycoords = np.array([[5, 5], [10, 10]])
        finder = DAOStarFinder(threshold=5.0, fwhm=3.0,
                               xycoords=xycoords)
        assert '<array; shape=(2, 2)>' in repr(finder)

    def test_threshold_2d_uniform(self, data):
        """
        Test that a uniform 2D threshold gives the same results
        as a scalar threshold.
        """
        threshold = 5.0
        fwhm = 1.0
        finder_scalar = DAOStarFinder(threshold, fwhm)
        finder_2d = DAOStarFinder(np.full(data.shape, threshold), fwhm)
        tbl_scalar = finder_scalar(data)
        tbl_2d = finder_2d(data)
        assert_array_equal(tbl_scalar, tbl_2d)

    def test_threshold_2d_varying(self, data):
        """
        Test that a varying 2D threshold detects fewer sources in
        regions with a higher threshold.
        """
        fwhm = 1.0
        threshold_low = 1.0
        threshold_high = 100.0
        threshold_2d = np.full(data.shape, threshold_low)
        threshold_2d[0:50, :] = threshold_high

        finder_low = DAOStarFinder(threshold_low, fwhm)
        finder_2d = DAOStarFinder(threshold_2d, fwhm)

        tbl_low = finder_low(data)
        tbl_2d = finder_2d(data)
        assert len(tbl_low) > len(tbl_2d)
        # All 2D sources should be in the lower half
        assert all(tbl_2d['y_centroid'] >= 50)

    def test_threshold_2d_repr(self):
        """
        Test repr with a 2D threshold array.
        """
        threshold = np.ones((10, 10))
        finder = DAOStarFinder(threshold=threshold, fwhm=3.0)
        assert '<array; shape=(10, 10)>' in repr(finder)
        assert '<array; shape=(10, 10)>' in str(finder)

    def test_threshold_2d_with_units(self, data):
        """
        Test that a 2D threshold with units works correctly.
        """
        units = u.Jy
        threshold = 5.0
        fwhm = 1.0
        threshold_2d = np.full(data.shape, threshold) * units
        finder = DAOStarFinder(threshold_2d, fwhm)
        tbl = finder(data << units)
        assert len(tbl) > 0

    def test_scale_threshold_default(self, data):
        """
        Test that scale_threshold=True (default) applies rel_err
        scaling.
        """
        threshold = 5.0
        fwhm = 1.5
        finder_default = DAOStarFinder(threshold, fwhm)
        finder_explicit = DAOStarFinder(threshold, fwhm,
                                        scale_threshold=True)
        tbl_default = finder_default(data)
        tbl_explicit = finder_explicit(data)
        assert_array_equal(tbl_default, tbl_explicit)
        # Verify the effective threshold is scaled
        assert finder_default.threshold_eff != threshold

    def test_scale_threshold_false(self, data):
        """
        Test that scale_threshold=False uses the threshold directly.
        """
        threshold = 5.0
        fwhm = 1.5
        finder = DAOStarFinder(threshold, fwhm, scale_threshold=False)
        assert finder.threshold_eff == threshold
        tbl = finder(data)
        assert len(tbl) > 0

    def test_scale_threshold_false_different_results(self, data):
        """
        Test that scale_threshold=False gives different results
        than the default.
        """
        threshold = 5.0
        fwhm = 1.5
        finder_scaled = DAOStarFinder(threshold, fwhm,
                                      scale_threshold=True)
        finder_unscaled = DAOStarFinder(threshold, fwhm,
                                        scale_threshold=False)
        tbl_scaled = finder_scaled(data)
        tbl_unscaled = finder_unscaled(data)
        # Different numbers of sources because effective thresholds
        # differ
        assert len(tbl_scaled) != len(tbl_unscaled)

    def test_scale_threshold_false_with_2d(self, data):
        """
        Test that scale_threshold=False works with a 2D threshold array.
        """
        fwhm = 1.5
        threshold_2d = np.full(data.shape, 5.0)
        finder = DAOStarFinder(threshold_2d, fwhm, scale_threshold=False)
        tbl = finder(data)
        assert len(tbl) > 0

    def test_scale_threshold_in_repr(self):
        """
        Test that scale_threshold appears in repr.
        """
        finder = DAOStarFinder(threshold=5.0, fwhm=3.0,
                               scale_threshold=False)
        assert 'scale_threshold=False' in repr(finder)
        assert 'scale_threshold: False' in str(finder)

    def test_deprecated_sharplo_sharphi(self):
        """
        Test that the deprecated 'sharplo'/'sharphi' keywords raise a
        warning and still work.
        """
        match = "The 'sharplo' and 'sharphi' parameters are deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            finder = DAOStarFinder(threshold=5.0, fwhm=3.0, sharplo=0.1)
        assert finder.sharpness_range == (0.1, 1.0)

        with pytest.warns(AstropyDeprecationWarning, match=match):
            finder = DAOStarFinder(threshold=5.0, fwhm=3.0, sharphi=2.0)
        assert finder.sharpness_range == (0.2, 2.0)

        with pytest.warns(AstropyDeprecationWarning, match=match):
            finder = DAOStarFinder(threshold=5.0, fwhm=3.0,
                                   sharplo=0.1, sharphi=2.0)
        assert finder.sharpness_range == (0.1, 2.0)

    def test_deprecated_roundlo_roundhi(self):
        """
        Test that the deprecated 'roundlo'/'roundhi' keywords raise a
        warning and still work.
        """
        match = "The 'roundlo' and 'roundhi' parameters are deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            finder = DAOStarFinder(threshold=5.0, fwhm=3.0, roundlo=-0.5)
        assert finder.roundness_range == (-0.5, 1.0)

        with pytest.warns(AstropyDeprecationWarning, match=match):
            finder = DAOStarFinder(threshold=5.0, fwhm=3.0, roundhi=0.5)
        assert finder.roundness_range == (-1.0, 0.5)

        with pytest.warns(AstropyDeprecationWarning, match=match):
            finder = DAOStarFinder(threshold=5.0, fwhm=3.0,
                                   roundlo=-0.5, roundhi=0.5)
        assert finder.roundness_range == (-0.5, 0.5)

    def test_deprecated_brightest(self):
        """
        Test that the deprecated 'brightest' keyword raises a warning
        and still works.
        """
        match = "'brightest' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            finder = DAOStarFinder(threshold=5.0, fwhm=3.0, brightest=5)
        assert finder.n_brightest == 5

    def test_deprecated_peakmax(self):
        """
        Test that the deprecated 'peakmax' keyword raises a warning
        and still works.
        """
        match = "'peakmax' was deprecated"
        with pytest.warns(AstropyDeprecationWarning, match=match):
            finder = DAOStarFinder(threshold=5.0, fwhm=3.0, peakmax=100.0)
        assert finder.peak_max == 100.0
